use core::{ops::Not, sync::atomic::Ordering};
use heapless::Vec;
use pico_website::unwrap;
use portable_atomic::{AtomicBool, AtomicU32};
use core::fmt::Write;

use crate::socket::HttpResCode;

static TURN: AtomicBool = AtomicBool::new(false);
// bits [0; 8] : player zero board / bits [9; 17] : player one board
static BOARD: AtomicU32 = AtomicU32::new(0);

pub struct GameClient {
    res_buf: Vec<u8, 4096>,
    team: Team
}
impl GameClient {
    pub fn new(team: Team) -> Self {
        Self {
            res_buf: Vec::new(),
            team
        }
    }
    pub async fn handle_request<'a>(&'a mut self, path: &str) -> (HttpResCode, &'static str, &'a [u8]) {
        if (path.starts_with("/ttt/cell") && path.len() == 10) || path == "/ttt/board" {
            let mut board = BOARD.load(Ordering::Acquire);
            let mut turn = TURN.load(Ordering::Acquire);

            // just return correct board in case of unauthorized move
            if path.starts_with("/ttt/cell") && self.team == turn.into() {
                let clicked_c: Cell = match TryInto::<Cell>::try_into(
                    unwrap(path.chars().nth(9).ok_or("no 9th char")).await,
                ) {
                    Ok(c) => c,
                    Err(_) => return (HttpResCode::NotFound, "", &[]),
                };
                if board
                    & ((2_u32.pow(clicked_c as u32))
                        + (2_u32.pow(9 + clicked_c as u32)))
                    != 0
                {
                    return (HttpResCode::Forbidden, "", &[]);
                }
                board = board | 2_u32.pow((self.team as u32 * 9) + clicked_c as u32);
                turn = (!self.team).into();
                BOARD.store(board, Ordering::Release);
                TURN.store(turn, Ordering::Release);
            }

            self.res_buf.clear();
            for c in 0..=8 {
                let picked_by = if board & 2_u32.pow(c) != 0 {
                    Some(Team::Zero)
                } else if board & 2_u32.pow(9 + c) != 0 {
                    Some(Team::One)
                } else {
                    None
                };
                match picked_by {
                    Some(Team::Zero) => {
                        unwrap(self.res_buf
                            .extend_from_slice(
                                b"<div class=\"cell\" style=\"background-color:blue\"></div>",
                            )).await;
                    }
                    Some(Team::One) => {
                        unwrap(self.res_buf.extend_from_slice(
                                b"<div class=\"cell\" style=\"background-color:red\"></div>",
                            )).await;
                    }
                    None => if self.team == turn.into() {
                        unwrap(write!(
                            self.res_buf,
                            "<button class=\"cell\" hx-post=\"/ttt/cell{}\" hx-trigger=\"click\" hx-target=\"#grid\" hx-swap=\"innerHTML\"></button>",
                            c
                        )).await;
                    } else {
                        unwrap(self.res_buf.extend_from_slice(
                                b"<div class=\"cell\"></div>",
                            )).await;
                    }
                };
            }
            (HttpResCode::Ok, "html", &self.res_buf)
        } else {
            (HttpResCode::NotFound, "", &[])
        }
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum Team {
    Zero = 0,
    One = 1,
}
impl From<bool> for Team {
    fn from(value: bool) -> Self {
        if value { Team::One } else { Team::Zero }
    }
}
impl Into<bool> for Team {
    fn into(self) -> bool {
        match self {
            Team::Zero => false,
            Team::One => true
        }
    }
}
impl Not for Team {
    type Output = Team;
    fn not(self) -> Self::Output {
        match self {
            Team::Zero => Team::One,
            Team::One => Team::Zero
        }
    }
}

#[derive(Debug, Clone, Copy)]
enum Cell {
    C0 = 0,
    C1 = 1,
    C2 = 2,
    C3 = 3,
    C4 = 4,
    C5 = 5,
    C6 = 6,
    C7 = 7,
    C8 = 8,
}
impl TryFrom<char> for Cell {
    type Error = ();
    fn try_from(value: char) -> Result<Self, Self::Error> {
        Ok(match value {
            '0' => Cell::C0,
            '1' => Cell::C1,
            '2' => Cell::C2,
            '3' => Cell::C3,
            '4' => Cell::C4,
            '5' => Cell::C5,
            '6' => Cell::C6,
            '7' => Cell::C7,
            '8' => Cell::C8,
            _ => return Err(()),
        })
    }
}
impl TryFrom<u8> for Cell {
    type Error = ();
    fn try_from(value: u8) -> Result<Self, Self::Error> {
        Ok(match value {
            0 => Cell::C0,
            1 => Cell::C1,
            2 => Cell::C2,
            3 => Cell::C3,
            4 => Cell::C4,
            5 => Cell::C5,
            6 => Cell::C6,
            7 => Cell::C7,
            8 => Cell::C8,
            _ => return Err(()),
        })
    }
}